'use strict';

const HttpStatus = require('http-status-codes');
const ErrorCodes = require('../libs/error-codes');

const DEFAULT_MESSAGE = 'A problem has occurred. If you continue to receive this message, contact support.';

class CpcError extends Error {
    constructor(orig) {
        super(
          typeof orig === 'string'
          ? orig
          : orig.userMessage || DEFAULT_MESSAGE
        );
        this.userMessage = this.message;
        this.systemMessage = orig.systemMessage || orig.message;
        this.statusCode = orig.statusCode || HttpStatus.INTERNAL_SERVER_ERROR;
        this.errorCode = orig.errorCode || null;
        this.title = orig.title || 'Whoops!';
        this.data = orig.data;
        this.originalError = orig;
    }
}

function formatError(error) {
    return formatErrorWithDefaultMessage(error, DEFAULT_MESSAGE);
}

function formatErrorWithDefaultMessage(error, defaultMessage = DEFAULT_MESSAGE) {
    let theError = error;
    if (error.originalError) {
        let originalError = error.originalError;
        while (originalError.errors && originalError.errors.length > 0) {
            originalError = originalError.errors[0];
            if (originalError.originalError) {
                originalError = originalError.originalError;
            }
        }

        theError = {
            title: originalError.title || 'Whoops!',
            message: originalError.userMessage || defaultMessage,
            userMessage: originalError.userMessage,
            systemMessage: originalError.systemMessage || error.message,
            statusCode: originalError.statusCode,
            errorCode: originalError.errorCode || null,
            data: originalError.data || null,
            locations: error.locations,
            path: error.path
        };

        if (process.env.NODE_ENV !== 'production') {
            let findRoot = error =>
                error.originalError ? findRoot(error.originalError) : error;
            let {stack, message} = findRoot(originalError);
            if (stack && stack.split) {
                stack = stack.split('\n').map(s => s.trim());
                theError.originalError = {message, stack};
            }
        }
    }
    return theError;
}

function badRequest(message) {
    const userMessage = message || DEFAULT_MESSAGE;
    return new CpcError({
        statusCode: HttpStatus.BAD_REQUEST,
        userMessage: userMessage
    });
}

function unAuthorized(message) {
    const userMessage = message || 'You are not Authorized to perform this request';
    return new CpcError({
        statusCode: HttpStatus.UNAUTHORIZED,
        userMessage: userMessage
    });
}

function notFound() {
    return new CpcError({
        statusCode: HttpStatus.NOT_FOUND,
        userMessage: 'Record not found'
    });
}

module.exports = {
    badRequest,
    unAuthorized,
    notFound,
    StatusCodes: HttpStatus,
    ErrorCodes,
    formatError,
    formatErrorWithDefaultMessage,
    CpcError
};
