const {CpcError} = require('./errorFormatter');
const logger = require('../libs/logger').getLogger(module);
const {makeExecutableSchema} = require('graphql-tools');
const path = require('path');
const {fileLoader, mergeTypes, mergeResolvers} = require('merge-graphql-schemas');
const transforms = require('./transforms');

async function resolverFunction(func, ...params) {
    try {
        if (!params) {
            logger.debug('Calling %s() with no params', func.name);
        } else {
            logger.debug('Calling %s with %d params', func.name, params.length);
        }
    } catch (error) {
        logger.debug('resolverFunction(): problem outputting function call: %s', error);
    }
    try {
        return await func.apply(null, params);
    } catch (error) {
        logger.error('Error running graphql resolver: %s', error);
        if (error.stack && (!error.statusCode || error.statusCode >= 500))
            logger.error(error.stack);
        if (error instanceof CpcError) throw error;
        throw new CpcError(error);
    }
}

/**
 * Add fields from data source to graphql resolver instance
 * @param instance
 * @param data
 */
function hydrateFields(instance, data) {
    Object.keys(data).forEach((key) => {
        instance[key] = data[key];
    });
}

/**
 * Generate an executable schema starting at the schemaPath root directory.
 * @param schemaPath
 */
function getExecutableSchema(schemaPath) {
    let resolversArray = fileLoader(path.join(schemaPath, '**/*Resolver.js'), {recursive: true});
    let typesArray = fileLoader(path.join(schemaPath, './**/*Schema.graphql'), {recursive: true});

    let mergedResolvers = mergeResolvers(resolversArray);
    let mergedTypes = mergeTypes(typesArray);

    return makeExecutableSchema({
        typeDefs: mergedTypes,
        resolvers: mergedResolvers
    });
}

module.exports = {
    hydrateFields,
    resolverFunction,
    getExecutableSchema,
    CpcError,
    transforms
};
