const TrieSearch = require('trie-search');
const logger = require('./logger').getLogger(module);

class Blacklist {
    constructor(words, defaultTransforms) {
        this.ts = new TrieSearch('word', {min: Blacklist.MIN_SEARCH_LEN});
        this.ts.addAll(words.map(word => ({
            word: word.replace(/\s/g, '').toLowerCase()
        })));
        this.defaultTransforms = defaultTransforms;
        this.search = this.search.bind(this);
        this.isBlacklisted = this.isBlacklisted.bind(this);
        this.searchWithTransform = this.applyTransform.bind(this);
    }

    search(word) {
        return this.ts.get(word);
    }

    searchPartial(word) {
        let start = 0;
        let end = Blacklist.MIN_SEARCH_LEN;
        let match = '';
        while (end <= word.length) {
            const slice = word.substr(start, end - start);
            const results = this.ts.get(slice);

            if (results.length && slice === results[0].word) {
                match = results[0].word;
                break;
            } else if (!results.length || end === word.length) {
                // If there are no results or more letters to check, reset
                start++;
                end = start + Blacklist.MIN_SEARCH_LEN;
                continue;
            }
            end++;
        }
        logger.info(`blacklist word found: ${match}`);
        return match;
    }

    isBlacklisted(word, transforms) {
        transforms = transforms || this.defaultTransforms;
        let query = word;
        for (const transform of transforms) {
            query = this.applyTransform(query, transform)
        }
        const result = this.search(query)[0];
        // if exact match isn't found, check for partial match
        const match = !!result && query === result.word
            || this.searchPartial(query);
        return !!match;
    }

    applyTransform(word, transform) {
        let transformed;
        switch (transform) {
            case Blacklist.TRANSFORMS.LOWERCASE:
                transformed = word.toLowerCase();
                break;
            case Blacklist.TRANSFORMS.TRIM:
                transformed = word.trim();
                break;
            case Blacklist.TRANSFORMS.ALPHA:
                const match = word.match(/[A-z]/g);
                if (!match) {
                    return '';
                }
                transformed = match.join('');
                break;
            case Blacklist.STRIP_WHITESPACE:
                transformed = word.replace(/\s/g, '');
                break;
        }
        return transformed;
    }
}

Blacklist.MIN_SEARCH_LEN = 3;

Blacklist.TRANSFORMS = {
    LOWERCASE: 'lowercase',
    TRIM: 'trim',
    ALPHA: 'alpha'
};

const defaultTransforms = [
    Blacklist.TRANSFORMS.LOWERCASE,
    Blacklist.TRANSFORMS.TRIM,
    Blacklist.TRANSFORMS.ALPHA
];

const words = [
    'acrotomophilia',
    'alaskan pipeline',
    'anilingus',
    'apeshit',
    'arsehole',
    'asshole',
    'assmunch',
    'autoerotic',
    'babeland',
    'bangbros',
    'bareback',
    'barely legal',
    'barenaked',
    'bastard',
    'bastardo',
    'bastinado',
    'bbw',
    'bdsm',
    'bestiality',
    'bimbos',
    'birdlock',
    'bitches',
    'blowjob',
    'blumpkin',
    'bollocks',
    'bondage',
    'brunette action',
    'bukkake',
    'bulldyke',
    'bullshit',
    'bunghole',
    'buttcheeks',
    'butthole',
    'camgirl',
    'camslut',
    'camwhore',
    'carpetmuncher',
    'chocolate rosebuds',
    'circlejerk',
    'cleveland steamer',
    'clitoris',
    'clusterfuck',
    'coprolagnia',
    'coprophilia',
    'cornhole',
    'creampie',
    'cunnilingus',
    'darkie',
    'daterape',
    'deepthroat',
    'dendrophilia',
    'dickhead',
    'dildo',
    'dingleberry',
    'dingleberries',
    'dirty pillows',
    'dirty sanchez',
    'doggiestyle',
    'doggystyle',
    'domination',
    'dominatrix',
    'double penetration',
    'dvda',
    'ejaculation',
    'erotic',
    'erotism',
    'escort',
    'faggot',
    'fellatio',
    'feltch',
    'female squirting',
    'femdom',
    'figging',
    'fingerbang',
    'fingering',
    'fisting',
    'footjob',
    'frotting',
    'fuck',
    'fuckin',
    'fucking',
    'fucktards',
    'fudgepacker',
    'futanari',
    'genitals',
    'girls gone wild',
    'goatcx',
    'goatse',
    'gokkun',
    'golden shower',
    'goodpoop',
    'goregasm',
    'g-spot',
    'handjob',
    'hardcore',
    'hentai',
    'homoerotic',
    'honkey',
    'how to murder',
    'huge fat',
    'humping',
    'incest',
    'jail bait',
    'jailbait',
    'jigaboo',
    'jiggaboo',
    'jiggerboo',
    'kinbaku',
    'kinkster',
    'kinky',
    'knobbing',
    'leather restraint',
    'leather straight jacket',
    'lolita',
    'lovemaking',
    'masturbate',
    'missionary position',
    'motherfucker',
    'muffdiving',
    'neonazi',
    'nigger',
    'nimphomania',
    'nipple',
    'nipples',
    'nsfw images',
    'nudity',
    'nympho',
    'nymphomania',
    'octopussy',
    'omorashi',
    'orgasm',
    'paedophile',
    'panties',
    'pedobear',
    'pedophile',
    'pegging',
    'pissing',
    'pisspig',
    'playboy',
    'ponyplay',
    'poontang',
    'punany',
    'poopchute',
    'porno',
    'pornography',
    'pthc',
    'pubes',
    'pussy',
    'queaf',
    'queef',
    'raghead',
    'raging boner',
    'rectum',
    'rimjob',
    'sadism',
    'santorum',
    'schlong',
    'scissoring',
    'shaved pussy',
    'shemale',
    'shibari',
    'shitblimp',
    'shota',
    'slanteye',
    's&m',
    'snatch',
    'sodomize',
    'sodomy',
    'splooge',
    'spooge',
    'strapon',
    'strappado',
    'swastika',
    'threesome',
    'throating',
    'titties',
    'topless',
    'towelhead',
    'tranny',
    'tribadism',
    'tubgirl',
    'tushy',
    'twat',
    'undressing',
    'upskirt',
    'urethra play',
    'urophilia',
    'vagina',
    'vibrator',
    'vorarephilia',
    'voyeur',
    'vulva',
    'wetback',
    'zoophilia'
];
module.exports = new Blacklist(words, defaultTransforms);
