const {addMiddleware} = require('graphql-add-middleware');

// Map of roles to handlers
const roles = {};

// Map of roles to child role precedence 1 depth deep
// All roles in the hierachy must be in the same chain
const hierachy = {};

let staticRoleResolver = function () {
    throw new Error('No static role resolver was set.');
};

/**
 * Handler for all static roles. Invokes the given `staticRoleResolver` function to
 * get the list of roles and returns wheter or not the user has the target role or a
 * role that inherits from it.
 *
 * @param {String} targetRole The name of the role with access
 */
function staticRoleHandler(targetRole) {
    return async function (root, args, context, info) {
        const userRoles = await staticRoleResolver(root, args, context, info) || [];
        return userRoles.some((userRole) => {
            return userRole === targetRole || inheritsFromRole(userRole, targetRole);
        });
    }
}

/**
 * Add new staic role definition. The role will be retrieved using staticRoleResolver
 *
 * @param {String} role The name of the role
 */
function addStaticRole(role) {
    roles[role] = staticRoleHandler(role);
}

/**
 * Add new dynamic role definition. Unlike static roles such as 'admin',
 * a dynamic role can change based on the result of handler.
 *
 * @param {String} role The name of the role
 * @param {Function} handler The role requirements
 */
function addRole(role, handler) {
    roles[role] = handler;
}


/**
 * Add __graphql_rbac_authed__ flag to context to denote the field is using auth middelware.
 * This is used to deny access to fields with no auth middelware set.
 */
async function markAuthed(root, args, context, info) {
    Object.defineProperty(context, '__graphql_rbac_authed__', {
        enumerable: false,
        configurable: false,
        writable: false,
        value: true
    });
}

/**
 * Combines all checkes to run concurrently
 *
 * @param {Array} handlers array of auth middleware functions
 */
function combineHandlers(handlers) {
    return async function (root, args, context, info, next) {
        const results = await Promise.all(handlers.map((handler) => {
            return handler(root, args, context, info);
        }));
        if (!results.some((success) => success === true)) {
            throw new Error('Unauthorized Access');
        }
        return next();
    }
}

/**
 * Assign authorization middleware to graphql schema fields
 *
 * @param {Object} schema GraphQL schema
 * @param {Object} config of fields to permissioned roles
 */
function init(schema, config) {
    // Parse config
    for (const type in config) { // ["Query", "Mutation"]
        const typeConfig = config[type];

        for (const field in typeConfig) {
            const fieldConfig = typeConfig[field];
            const configRoles = Object.keys(fieldConfig);
            const handlers = configRoles.filter(role =>{
                if (!roles[role]) {
                    throw new Error(`Invalid role '${role}' given`);
                }
                return true;
            }).map(role => roles[role]);
            const combinedHandlers = combineHandlers([markAuthed, ...handlers]);
            // Add auth middleware to graphql queries and mutations
            addMiddleware(schema, `${type}.${field}`, combinedHandlers);
        };
    };

    // Catch-all preventing access to fields in the schema without authorization configured
    // TODO: uncomment once all existing fields have auth configured
    // addMiddleware(schema, async function(root, args, context, info, next) {
    // 	if (!context.__graphql_rbac_authed__) {
    // 		throw new Error(`Authorization not set for field '${info.fieldName}'`);
    // 	}
    // 	return next();
    // });
}

/**
 * Allows roles to inherit permissions from another role
 *
 * @param {Array} roles List of roles in ascending permission order
 */
function setRoleHierachy(roles) {
    for (let i = roles.length - 1; i > 0; i--) {
        hierachy[roles[i]] = roles[i - 1];
    }
}

/**
 * Checks if one role inherits permissions from another role
 *
 * @param {String} role Top level role
 * @param {String} targetRole Lower level role
 * @returns {Boolean}
 */
function inheritsFromRole(role, targetRole) {
    while (hierachy[role]) {
        role = hierachy[role];
        if (role === targetRole) {
            return true;
        }
    }
    return false;
}

/**
 * Provides a setter for the function that resolves static user roles from requests
 *
 * @param {Function} fn async or promise returning function that resolves the requesting user's roles
 */
function setStaticRoleResolver(fn) {
    staticRoleResolver = fn;
}

module.exports = {
    init,
    addStaticRole,
    addRole,
    setRoleHierachy,
    setStaticRoleResolver
};
