const {PostgresPubSub} = require('graphql-postgres-subscriptions');
const {Client} = require('pg');
const uuidv4 = require('uuid/v4');

const ANNOUNCE_INTERVAL = 1000 * 30; // ms
const REFRESH_INTERVAL = 1000 * 60 * 15; // ms

module.exports = async function (app, clientUrl, publisher = true, notificationTopics) {
    // if publisher is set to true this messagebus will only publish messages
    // not deal with subscriptions
    const config = app.config;

    const TOPIC_KEYS = {};
    for (const key in notificationTopics) {
        const code = notificationTopics[key];
        TOPIC_KEYS[code] = key;
    }

    function getTopicKey(code) {
        return TOPIC_KEYS[code];
    }

    const logger = Object.create(app.Logger.getLogger(module));
    // Replace logger.debug() with no-op if so indicated by config
    // if (!config.bus.log) logger.debug = () => {};

    const client = new Client(clientUrl);
    const consumerGroup = config.app.serviceName;
    const subscriberUUID = uuidv4();

    const directedTopics = new Set();
    let live = false;
    let periodicQueries = [];
    let pubSub;

    async function announceListeners() {
        if (!live) {
            logger.debug('[announceListeners] Bus is not live');
            return;
        }

        if (directedTopics.size === 0) {
          // if we do not have topics, lets announce any
          return
        }

        const updateListenersSQL = `
            INSERT INTO message_bus_subscriptions(topic)
            VALUES ${[...directedTopics].map(topic => `('${topic}')`)}
            ON CONFLICT (topic) DO UPDATE SET updated_at = CURRENT_TIMESTAMP;
        `;

        await client.query(updateListenersSQL);
        logger.debug(
            {label: 'announceListeners'},
            'Updated Status: %j',
            {subscriberUUID, consumerGroup}
        );
    }

    function asyncIterator(...args) {
        if (!live) throw new Error('Bus is not live');
        return pubSub.asyncIterator(...args);
    }

    async function ensureSubscriptionsTable() {
        // This one we call while live = false still
        return client.query(`
          CREATE TABLE IF NOT EXISTS "message_bus_subscriptions" (
            "id"         SERIAL PRIMARY KEY,
            "topic"      varchar(63) UNIQUE NOT NULL,
            "updated_at" TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP
          );
        `);
    }

    async function publish(topic, message, {broadcast = false} = {}) {
        if (!live) return false;

        if (!broadcast) {
            const destinations = await getUniqueGroupConsumers(topic);

            logger.debug(
                {label: 'publish'},
                'Publishing Message: %j',
                {topic: getTopicKey(topic), destinations}
            );

            if (!destinations) return false;
            for (const destination of destinations) {
                logger.debug(
                    {label: 'publish'},
                    'Non-Broadcast Destination: %j',
                    destination
                );
                await publishOnce(destination, message);
            }
        } else {
            logger.debug({label: 'publish'}, 'Broadcast Topic: %j', topic);
            await publishOnce(topic, message);
        }

        return true;
    }

    async function publishOnce(topic, message) {
        if (!topic) throw new Error(`Invalid topic: ${topic}`);
        try {
            await pubSub.publish(topic, message);
        } catch (err) {
            logger.error('Publish to %s failed: %s, payload: %j', topic, err, message);
            throw err;
        }
    }

    async function refreshListeners() {
        if (!live) {
            logger.debug({label: 'refreshListeners'}, 'Bus is not live');
            return;
        }

        const truncateListenersSQL = `
          DELETE
          FROM message_bus_subscriptions
          WHERE updated_at < CURRENT_TIMESTAMP - interval '30 seconds';
        `;
        await client.query(truncateListenersSQL);
        logger.debug(
            {label: 'refreshListeners'},
            'Refreshed Listeners: %j',
            {subscriberUUID, consumerGroup}
        );
    }

    async function start() {
        // PubSub ctor calls, but does not await, connect(), so let's do that
        // ourselves to be sure.
        logger.debug('connecting')
        await client.connect(() => logger.debug('connected'))

        pubSub = new PostgresPubSub({client});
        await ensureSubscriptionsTable();

        live = true;

        periodicQueries.push(setInterval(refreshListeners, REFRESH_INTERVAL));
        periodicQueries.push(setInterval(announceListeners, ANNOUNCE_INTERVAL));
    }

    async function stop() {
        live = false;
        periodicQueries.forEach(clearInterval);
        await client.end();
    }

    async function subscribe(topic, callback) {
        // if we are setup as publishers, we do not subscribe to topics.
        if (publisher === true) {
          return
        }

        if (!live) throw new Error('Bus is not live');
        if (!topic) throw new Error(`Invalid topic: '${topic}'`);

        const directedTopic = `${topic}/${consumerGroup}/${subscriberUUID}`;
        if (directedTopic.length > 63) {
            throw new Error(`Topic '${directedTopic}' is too long`);
        }

        const meta = {topic: getTopicKey(topic), subscriberUUID, consumerGroup};

        const wrapper = (...args) => {
            logger.debug(
                {label: 'subscribe'},
                'Received Message: %j',
                {...meta, data: args[0]}
            );
            return callback.apply(this, [...args]);
        };

        logger.debug({label: 'subscribe'}, 'Subscribed: %j', meta);

        const [globalCounter, localCounter] = await Promise.all([
            pubSub.subscribe(topic, wrapper),
            pubSub.subscribe(directedTopic, wrapper)
        ]);

        directedTopics.add(directedTopic);

        try {
            await announceListeners();
        } catch (error) {
            logger.error({label: 'announceListeners'}, '%j', error);
        }

        return {globalCounter, localCounter};
    }

    async function getUniqueGroupConsumers(topic) {
        const sql = `
            SELECT topic
            FROM message_bus_subscriptions
            WHERE topic like '${topic}/%'
            AND updated_at > CURRENT_TIMESTAMP - INTERVAL '30 seconds';
        `;
        try {
            const destinations = [];
            const listeners = (await client.query(sql)).rows.map(row => {
                const [id, group, uuid] = row.topic.split('/');
                return {group, uuid};
            });
            const groups = new Map();
            for (const {group, uuid} of listeners) {
                const peers = groups.get(group) || [];
                peers.push(uuid);
                groups.set(group, peers);
            }
            for (const [group, peers] of groups) {
                const target = Math.round(Math.random() * (peers.length - 1));
                const peer = peers[target];
                destinations.push(`${topic}/${group}/${peer}`);
            }
            if (!destinations.length) {
                throw new Error(`No consumers available for '${topic}'`);
            }
            return destinations;
        } catch (err) {
            logger.error(err);
            return null;
        }
    }

    return {
        // Personally, I'd just put publish() and subscribe() directly on the
        // returned object, but we can do that later
        bus: {publish, subscribe, asyncIterator},
        start,
        stop
    };
};
