'use strict'
const { Storage } = require('@google-cloud/storage')
const path = require('path')
const url = require('url')

// TODO: enforce that some methods take URLs, and some paths?

const BASE_URL = {
  protocol: 'https',
  hostname: 'storage.googleapis.com'
}
const CACHE_CONTROL = 'public, max-age=86400'

module.exports = ({
  Logger: { getLogger },
  config: {
    bucket: bucketName,
    project: { id: projectId }
  }
}) => {
  const logger = getLogger(module)
  const bucket = new Storage({ projectId }).bucket(bucketName)

  function clearStorage() {
    throw new Error('Not implemented!')
  }

  /**
   * Returns a promise to delete the file with a given name and path returning a flag which is true if the file is
   * deleted successfully.
   *
   * @param {string} file The name and path of the file to be deleted.
   *
   * @returns {Promise<boolean>} The promise promise to delete the file with the given name and path returning a flag
   * which is true if the file is deleted successfully.
   */
  async function deleteUrl(file) {
    if (!file || file === 'none') return true
    await getFile(file).delete()
    return true
  }

  /**
   * Returns a link to the file with a given name and path in the current bucket.
   *
   * @param {string} urlOrPath The name and path or URL of the file.
   *
   * @returns {File} The file with the given name and path in the current bucket.
   */
  function getFile(urlOrPath) {
    return bucket.file(isUrl(urlOrPath) ? urlToPath(urlOrPath) : urlOrPath)
  }

  function getInfo(file) {
    throw new Error('Not implemented!')
  }

  /**
   * Returns a promise containing a secure URL created from a given original URL with a given expiration in minutes.
   *
   * @param {string} url The given URL.
   *
   * @param {number} expirationMinutes The expiration in mutes.
   *
   * @returns {string} A promise containing the secure URL created from the given original URL.
   */
  async function getSecureUrl(file, expirationMinutes = 1) {
    if (!file) return null
    return (await getFile(file).getSignedUrl({
      action: 'read',
      expires: Date.now() + expirationMinutes * 60000
    }))[0]
  }

  function getStream(file) {
    return getFile(file).createReadStream()
  }

  function isUrl(str) {
    return Boolean(url.parse(str).protocol)
  }

  function pathToUrl(p) {
    return url.format({
      ...BASE_URL,
      pathname: path.join(`${bucketName}`, p)
    })
  }

  function urlToPath(u) {
    let { pathname } = url.parse(u)

    try {
      // urls are stored in the database
      // to convert them to filenames they have to be decoded
      pathname = decodeURIComponent(pathname)
    } catch (e) {
      // no problem if it fails
    }

    // Also trims off the bucket name
    // slice(2) because pathname's leading / becomes the empty string
    return pathname.split(path.sep).slice(2).join(path.sep)
  }

  /**
   * Returns a promise containing the response to a request to store a given file stream as a file with a given name
   * and path
   *
   * @param {NodeJS.ReadStream} origin The stream to be saved.
   *
   * @param {string} destinationPath The path to the location where the stream is to be saved.
   *
   * @param {boolean} signUrl A flag whuch is true if the file URL must be saved.
   *
   * @param {number} expirationMinutes The expiration, in minutes, of the URL signature.
   *
   * @returns {Promise<*>} The promise containing the response to a request to store the given file stream as a file
   * with the given name and path.
   */
  function saveStream(
    origin,
    destinationPath,
    { signUrl = false, expirationMinutes = 1, metadata , createWriteStreamOpts = {}} = {}
  ) {
    let file = bucket.file(destinationPath)
    let destination = file.createWriteStream({ metadata, ...createWriteStreamOpts });
    let uploadUrl = pathToUrl(file.name)
    return new Promise((resolve, reject) => {
      destination.on('finish', () => {
        logger.info(`Upload complete: ${uploadUrl}`)
        resolve(signUrl ? getSecureUrl(uploadUrl, expirationMinutes) : uploadUrl)
      })
      destination.on('error', reject)
      origin.pipe(destination)
    })
  }

  async function savePath(origin, destination) {
    logger.debug('Uploading photo to bucket: %s', bucketName)
    await bucket.upload(origin, {
      destination,
      public: false,
      resumable: false,
      gzip: false,
      metadata: {
        cacheControl: CACHE_CONTROL
      }
    })
    return pathToUrl(destination)
  }

  async function makePublic(filepath) {
    return bucket
      .file(filepath)
      .makePublic()
  }

  return {
    getSecureUrl,
    savePath,
    saveStream,
    deleteUrl,
    clearStorage,
    getInfo,
    getStream,
    pathToUrl,
    makePublic
  }
}
