'use strict'

const fs = require('fs-extra')
const path = require('path')
const url = require('url')

module.exports = async ({ Logger, tempDir }) => {
  let log = Logger.getLogger(module)

  let tempPath = path.join(tempDir.path, 'local-storage')
  await fs.mkdirp(tempPath)

  function isUrl(str) {
    return Boolean(url.parse(str).protocol)
  }

  function getFile(file) {
    return urlToPath(file)
  }

  function urlToPath(uri) {
    return uri.replace(/^file:\/\//, '')
  }

  function pathToUrl(path) {
    return 'file://' + path
  }

  // Wouldn't be hard to implement either this or getFileInfo(), but they're not
  // implemented for our Google storage handler, so it'd be misleading to
  // implement them here, for test purposes.
  function clearStorage() {
    let message = 'Local: clearStorage(): not implemented yet'
    log.error(message)
    throw new Error(message)
  }

  /**
   * Returns a promise to delete the file at a given URL.
   *
   * @param {string} url The URL of the file to be deleted.
   *
   * @returns {Promise<void>} The promise to delete the file at the given URL.
   */
  function deleteUrl(uri) {
    return fs.unlink(getFile(uri))
  }

  function getInfo(uri) {
    let message = 'Local: getFileInfo(): not implemented yet'
    log.error(message)
    throw new Error(message)
  }

  function getStream(uri) {
    return fs.createReadStream(getFile(uri))
  }

  /**
   * Save image to a local storage for testing
   * @param origin
   * @returns {Promise<String>}
   */
  async function savePath(origin, dest) {
    let fullDest = path.join(tempPath, path.basename(dest))
    await fs.copyFile(origin, fullDest)
    return pathToUrl(fullDest)
  }

  /**
   * Returns a promise containing the response to a request to store a given file stream as a file with a given name
   * and path
   *
   * @param {NodeJS.ReadStream} stream The stream to be saved.
   *
   * @param {string} destination The path to the location where the stream is to be saved.
   *
   * @returns {Promise<*>} The promise containing the response to a request to store the given file stream as a file
   * with the given name and path.
   */
  async function saveStream(stream, destination) {
    let fullPath = path.join(tempPath, destination)
    await fs.mkdirp(path.parse(fullPath).dir)

    let dest = fs.createWriteStream(fullPath)
    stream.on('error', err => dest.destroy(err))

    return new Promise((resolve, reject) => {
      dest.on('finish', () => resolve(pathToUrl(fullPath))).on('error', reject)
      stream.pipe(dest)
    })
  }

  /**
   * Returns a promise containing a secure URL created from a given original URL with a given expiration in minutes.
   *
   * @param {string} url The given URL.
   * @returns {string} The secure URL created from the given original URL. It's the local filesystem, so we just give a file URL back.
   */
  async function getSecureUrl(file) {
    return isUrl(file) ? file : pathToUrl(file)
  }

  async function makePublic(filepath) {
    // all local files are "public"
    return;
  }

  return {
    clearStorage,
    deleteUrl,
    getInfo,
    getStream,
    savePath,
    saveStream,
    getSecureUrl,
    pathToUrl,
    makePublic
  }
}
