module.exports = function (app) {
    const {Logger, config, dbUtils} = app;
    const logger = Logger.getLogger(module);
    const queryAsync = dbUtils.getQueryAsync(dbUtils.connector);
    const  {
        setStaticRoleResolver,
        addRole,
        addStaticRole,
        mapArgsToType,
        setRoleHierachy
    } = require('../libs/graphql-rbac');
    let {rules, schemaMap} = config.acl;
    schemaMap = applySchemaMapDefaults(schemaMap);

    function applySchemaMapDefaults(schemaMap) {
        const map = {};
        for (const schemaName in schemaMap) {
            const tableCols = schemaMap[schemaName];
            if (typeof tableCols === 'string') {
                map[schemaName] = { table: tableCols };
            } else {
                map[schemaName] = tableCols;
            }
            // Make deep copy and set defaults
            map[schemaName] = {
                idField: 'id',
                ownerField: 'userId',
                ...map[schemaName]
            };
            if (map[schemaName].join) {
                map[schemaName].join = {
                    idField: 'id',
                    ownerField: 'userid',
                    ...(map[schemaName].join || {})
                };
            }
        }
        return map;
    }

    async function verifyOwner(userId, args, config)  {
        const queryParams = [];
        args = flattenKeys(args);
        config = flattenKeys(config);

        for (const argName in args) {
            const idValue = args[argName];
            if (!config.hasOwnProperty(argName)) {
                continue; // config doesn't auth this arg
            }

            const schemaName = config[argName];
            if (schemaName === 'User') {
                if (idValue !== userId) {
                    return false;
                } else {
                    continue; // user verified from userId
                }
            }
            if (!schemaMap.hasOwnProperty(schemaName)) {
                throw new Error(`Unable to verify ownership of ${argName} <${schemaName}> by ${userId}`);
            }
            const {table, ownerField, idField, join} = schemaMap[schemaName];
            queryParams.push({table, idField, idValue, ownerField, userId, join});
        }
        return queryParams.length ? verifyByIds(queryParams) : true;
    }

    function getVerifyByIdSubQuery({table, idField, idValue, ownerField, userId, join}, i) {
        const idCheck = Array.isArray(idValue) ?
            `${idField} IN(${idValue.join()})` : `${idField} = ${idValue}`;

        const query = join ?
            `SELECT COUNT(DISTINCT table0.${idField})
            FROM ${table} table0
            RIGHT JOIN ${join.table} table1
            ON(table0.${join.associativeField} = table1.${join.idField} AND table1.${join.ownerField} = ${userId})
            WHERE table0.${idCheck}`
            :
            `SELECT COUNT(DISTINCT ${idField})
            FROM ${table} WHERE ${idCheck}
            AND ${ownerField} = ${userId}`;

        return `(${query}) as count${i}`;
    }

    async function verifyByIds(params) {
        // Use subquery to avoid multiple requests for each arg
        let numChecks = 0;
        const subQueries = params.map((params, i)=>{
            numChecks += Array.isArray(params.idValue) ? new Set(params.idValue).size : 1;
            return getVerifyByIdSubQuery(params, i);
        }).join(',\n');
        const query = `SELECT ${subQueries} LIMIT 1;`;

        const rows = await queryAsync(query);
        const result = rows[0];
        if (!result) {
            throw new Error('Unable to query for ownership');
        }
        const totalOwned = Object.keys(result).reduce((acc, col)=>acc + parseInt(result[col]), 0);
        return numChecks === totalOwned;
    }

    function getUserFromContext(context) {
        try {
            const {user} = context.graphqlContext;
            return user;
        } catch (err) {
            return null;
        }
    }

    function addDefaultRoles() {
        // Dynamic Roles
        addRole('everyone', async (root, args, context, info) => true);
        addRole('authenticated', async (root, args, context, info) => {
            const user = getUserFromContext(context);
            return user && !!user.token;
        });
        addRole('unauthenticated', async (root, args, context, info) => {
            const user = getUserFromContext(context);
            return !user || !user.token;
        });

        addRole('owner', async (root, args, context, info) => {
            const user = getUserFromContext(context);
            if (!user) {
                return false;
            }
            try {
                const operation = capitalize(info.operation.operation);
                const roleConfig = rules[operation][info.fieldName].owner;
                return await verifyOwner(user.id, args, roleConfig);
            } catch (err) {
                logger.error(err);
                return false;
            }
        });

        // Used to access roles from user
        const roleBoolMap = {
            isAdmin: 'admin',
            isSuperAdmin: 'superAdmin',
            isCanonRepresentative: 'canonRepresentative',
        };
        setStaticRoleResolver(async (root, args, context, info) => {
            const roles = [];
            const user = getUserFromContext(context);
            if (!user) {
                return roles;
            }
            for (const roleBool in roleBoolMap) {
                if (user[roleBool]) {
                    roles.push(roleBoolMap[roleBool]);
                }
            }
            return roles;
        });

        // Static Roles
        addStaticRole('client');
        addStaticRole('canonRepresentative');
        addStaticRole('admin');
        addStaticRole('superAdmin');

        setRoleHierachy([
            'everyone',
            'client',
            'canonRepresentative',
            'admin',
            'superAdmin'
        ]);
    }

    function capitalize(str) {
        const letters = str.split('');
        letters[0] = letters[0].toUpperCase();
        return letters.join('');
    }

    function flattenKeys(obj, parentKey='', sep='.') {
        const flatObj = {};
        for(let key in obj) {
            let value = obj[key];
            if (typeof value === 'object') {
                let {childKey, baseValue} = flattenKeys(value, key);
                key = `${key}${sep}${childKey}`;
                value = baseValue;
            }
            if (parentKey) {
                return {childKey: key, baseValue: value};
            } else  {
                flatObj[key] = value;
            }
        }
        return flatObj;
    }

    return {
        addDefaultRoles
    };
};