'use strict'
const dotenv = require('dotenv');
const fs = require('fs');
const path = require('path');

const logger = require('../libs/logger').getLogger(module);

function loadDotenv(nodeEnv, cwd = process.cwd()) {
    const ext = (nodeEnv === 'development' ? '' : `.${nodeEnv}`);
    const envFile = path.resolve(cwd, `.env${ext}`);
    if (fs.existsSync(envFile)) {
        dotenv.config({path: envFile});
    } else if (nodeEnv === 'test') {
        throw new Error(`${nodeEnv}-specific env file not found: ${envFile}`);
    }
    // or else we just exit silently, not requiring the file if it's not there
}

function getConfig(theModel) {
    return theModel.find();
}

// noinspection FunctionWithMultipleLoopsJS
async function createConfigInDb(theModel, theConfig) {
    let configKeys = convertConfigToKeys(theConfig);
    for (const key of configKeys) {
        const currentKey = await theModel.find({where: {configKey: key}});
        if (currentKey.length === 0) {
            const newKeys = key.split('.');
            let newConfig = theConfig;
            for (let i = 0; i < newKeys.length - 1; i++) {
                newConfig = newConfig[newKeys[i]];
            }

            let value = newConfig[newKeys[Math.max(0, newKeys.length - 1)]];
            // empty config values fail model validation
            if (value && value.length > 0) {
                logger.info(`Adding key ${key}`);
                await theModel.create({configKey: key, value});
            } else {
                logger.info(`Blank value: skipping key ${key}`);
            }
        }
    }
    return true;
}

function convertConfigToKeys(config) {
    let list = [];
    (function (o, r) {
        // noinspection AssignmentToFunctionParameterJS
        r = r || '';
        if (typeof o !== 'object') {
            return true;
        }
        if (Array.isArray(o)) {
            return true;
        }
        for (let c in o) {
            // noinspection JSUnfilteredForInLoop
            if (arguments.callee(o[c], r + '.' + c)) {
                list.push(r.substring(1) + '.' + c);
            }
        }
        return false;
    })(config);
    return list;
}

function convertJsonToArray(obj, parent = null) {
    let output = [];
    for (let key in obj) {
        let newKey = parent + '.' + key;
        if (obj.hasOwnProperty(key) && typeof obj[key] === 'object') {
            let results = (convertJsonToArray(obj[key], newKey));
            output.push(...results);
        } else {
            output.push({configKey: newKey, value: obj[key]});
        }
    }
    return output;
}

async function updateConfigFromDb(theModel, theConfig, messageBus = null) {
    const configs = await getConfig(theModel);
    for (let i = 0; i < configs.length; i++) {
        writeConfigKey(theConfig, configs[i].configKey, configs[i].value, configs[i].type);
    }
    if (messageBus) {
        messageBus.bus.publish(messageBus.TOPICS.CONFIG_UPDATED, null, {broadcast: true}).catch(error => logger.error('updateConfigFromDb(): %s', error));
    }
}

function writeConfigKey(theConfig, key, value, type) {
    try {
        const newKeys = key.split('.');

        let newConfig = theConfig;

        for (let i = 0; i < newKeys.length - 1; i++) {
            newConfig = newConfig[newKeys[i]];
        }
        let finalVal = value;
        if (type !== null && value !== null) {
            switch (type.toLowerCase()) {
                case 'boolean':
                    finalVal = (value.toLowerCase() === 'true');
                    break;
                case 'integer':
                    finalVal = parseInt(value);
                    break;
                case 'array':
                    finalVal = value.split(',');
                    break;
                case 'date':
                    finalVal = new Date(Date.parse(value));
                    break;
                default:
                    break;
            }
        }
        newConfig[newKeys[newKeys.length - 1]] = finalVal;
    } catch (error) {
        logger.debug('Problem writing config key: %s: %s, %s', key, value, error);
    }
}

async function writeKeyValue(theModel, key, value, type) {
    return await theModel.upsert({configKey: key, value: value, type: type});
}

async function writeKeyValues(theModel, inputs, configuration, name, messageBus = null) {
    let result = null;
    try {
        for (let i = 0; i < inputs.length; i++) {
            await writeKeyValue(theModel, inputs[i].configKey, inputs[i].value, inputs[i].type);
        }

        await updateConfigFromDb(theModel, configuration, messageBus);
        result = convertJsonToArray(configuration, name);
    } catch (error) {
        logger.error('Error updating config key/value: %s', error);
    }
    return result;
}

async function readJsonConfigFromDb(theModel, config) {
    const configs = await theModel.find();
    for (let aConfig of configs) {
        config[aConfig.key] = aConfig.value;
    }
    return config;
}

async function writeJsonConfigToDb(theModel, theConfig) {
    for (const key of Object.keys(theConfig)) {
        await theModel.upsert({key, value: theConfig[key]});
    }
}

async function writeJsonKeyValues(theModel, inputs, theConfig) {
    for (const input of inputs) {
        await theModel.upsert({key: input.key, value: JSON.parse(input.value)});
    }
    return readJsonConfigFromDb(theModel, theConfig);
}

module.exports = {
    loadDotenv,
    getConfig,
    writeKeyValue,
    writeKeyValues,
    convertJsonToArray,
    updateConfigFromDb,
    createConfigInDb,
    readJsonConfigFromDb,
    writeJsonConfigToDb
};
