const logger = require('./../libs/logger').getLogger(module);
const moment = require('moment');

function createPagingSql(paging) {
    let sql = '';
    if (paging) {
        if (paging.orderBy) {
            sql += ' ORDER BY ';
            let comma = false;
            for (let order of paging.orderBy) {
                if (comma) {
                    sql += ',';
                }
                comma = true;
                sql += order;
            }
        }
        if (paging.count) {
            sql += ` LIMIT ${paging.count} `;
        }

        if (paging.skip) {
            sql += ` OFFSET ${paging.skip} `;
        }
    }
    return sql;
}

function createPagingFilter(paging, where) {
    let newWhere = Object.assign({}, where);
    if (paging) {
        if (paging.count) {
            newWhere.limit = paging.count;
        }

        if (paging.skip) {
            newWhere.skip = paging.skip;
        }
        if (paging.orderBy) {
            newWhere.order = paging.orderBy;
        }
    }
    return newWhere;
}

async function cleanIt(theModel, filter) {
    if (filter) {
        const results = await theModel.destroyAll(filter);
        if (results.count > 0) {
            logger.debug('Cleaned up %d orphan %s', results.count, theModel.name);
        }
    }
}

function addUploadedFilter(where, uploadedOnly = false) {
    const filter = {url: {neq: null}};
    const originalWhere = where || {};
    const newWhere = Object.assign({}, originalWhere);
    if (uploadedOnly) {
        if (originalWhere.where) {
            if (!where.where.and) {
                newWhere.where = {and: [originalWhere.where]};
            }
            newWhere.where.and.push(filter);
        } else {
            newWhere.where = filter;
        }
    }
    return newWhere;
}

function getDateStartEnd(dateRange) {
    if (!dateRange) {
        return null;
    }
    let start = null;
    let end = new Date(Date.now());
    if (dateRange.relative) {
        if (dateRange.relative.daysBack) {
            start = moment().subtract(dateRange.relative.daysBack, 'days').toDate();
        } else if (dateRange.relative.weeksBack) {
            start = moment().subtract(dateRange.relative.weeksBack, 'weeks').toDate();
        } else if (dateRange.relative.monthsBack) {
            start = moment().subtract(dateRange.relative.monthsBack, 'months').toDate();
        } else if (dateRange.relative.yearsBack) {
            start = moment().subtract(dateRange.relative.yearsBack, 'years').toDate();
        } else {
            throw 'Invalid date range provided';
        }
    } else {
        start = new Date(dateRange.startDate);
        if (dateRange.end) {
            end = new Date(dateRange.endDate);
        }
    }
    return {start, end};
}

function createDateRangeFilter(dateRange = null, attribute = null, where = {}) {
    if (!dateRange) {
        return where;
    }
    const {start, end} = getDateStartEnd(dateRange);
    let startFilter = {};
    startFilter[attribute] = {gte: start};
    let endFilter = {};
    endFilter[attribute] = {lte: end};
    const newWhere = Object.assign({}, where);
    if (where.where) {
        if (!where.where.and) {
            newWhere.where = {and: [where.where]};
        }
    } else {
        newWhere.where = {and: []};
    }
    newWhere.where.and.push(startFilter);
    newWhere.where.and.push(endFilter);
    return newWhere;
}

function createDateRangeSql(dateRange = null, attribute) {
    if (!dateRange) {
        return '';
    }
    const {start, end} = getDateStartEnd(dateRange);
    return ` ${attribute} >= '${start.toISOString()}' and ${attribute} <= '${end.toISOString()}' `;
}

module.exports = {
    createPagingSql,
    createPagingFilter,
    createDateRangeFilter,
    createDateRangeSql,
    cleanIt,
    addUploadedFilter
};
