'use strict'
// See database/{create,update}.js in any of the services for examples of usage.

const {boundPromisify} = require('./miscUtils')

module.exports = app => {
    let {Logger} = app
    let log = Logger.getLogger({id: 'utils/schemaUtils.js'})

    /*
     * @see http://apidocs.strongloop.com/loopback-datasource-juggler/#datasource-prototype-automigrate
     */
    async function create(dataSource) {
        log.info(`Automigrating ${dataSource.name}`)
        await boundPromisify(dataSource, 'automigrate')()
    }

    // command param is either the create or update function in this file
    async function invoke(command) {
        let {config, dataSources} = app

        for (let name of config.db.datasources) {
            let dataSource = dataSources[name]

            if (!dataSource) throw new Error('Could not find data source!')

            log.info(`Using data source ${name} at ${redactPassword(dataSource.connector.clientConfig.url)}`)
            await command(dataSource)
        }
    }

    // Use this if you're running a schema change as a standalone script, and
    // therefore need to load and unload the app.
    async function main(command) {
        let exitCode = 0
        try {
            await app.load({initServer: false, scheduleJobs: false})
            await invoke(command)
        } catch (err) {
            log.error(err)
            exitCode = 1
        } finally {
            await app.unload()
        }
        return exitCode
    }

    // Keep our DB password out of the logs. On GCP the instance is IP-white-
    // listed so it's not too a big deal, but just in case.
    function redactPassword(url) {
        return url.replace(
            /(\w+:\/\/)((\w+)(:.*)?@)?(.+)/,
            (match, head, creds, user, pass, tail) => {
                if (creds) return `${head}${user}:[REDACTED]@${tail}`
                else return `${head}${tail}`
            }
        )
    }

    /*
     * @see http://apidocs.strongloop.com/loopback-datasource-juggler/#datasource-prototype-autoupdate
     */
    async function update(dataSource) {
        if (await boundPromisify(dataSource, 'isActual')()) {
            log.info(`Nothing to update in ${dataSource.name}`)
            return
        }

        log.info(`Changes detected, updating ${dataSource.name}`)
        await boundPromisify(dataSource, 'autoupdate')()
    }

    return { create, invoke, main, update }
}
