'use strict'
const logger = require('../libs/logger').getLogger(module)
const cron = require('node-schedule')
const path = require('path')
const configUtils = require('../utils/configUtils')
const dbUtils = require('../utils/dbUtils')

const JOB_FILE = 'jobs/jobsToRun.js'
const POD_TO_RUN = 'podToRun'
const jobList = new Map()

function cancelJobs() {
    logger.debug('Cancelling jobs...')
    Object.values(cron.scheduledJobs).forEach(j => {
        logger.debug(`Cancelling job ${j.name}`)
        // Not actually async, because we're just removing it from the schedule
        j.cancel()
    })
}

async function initServer({server, config, configModel}) {
    await configUtils.updateConfigFromDb(configModel, config)
    try {
        logger.debug('Updating database')
        await dbUtils.dbUpdate(server, config, configModel)
    } catch (error) {
        logger.error('Boot error: %s', error)
        throw error
    }
}

function getTheJob(jobFile, config, configModel, key) {
    return async function () {
        await configUtils.updateConfigFromDb(configModel, config) // need to load to see if we have an new pod to run
        const jobConfig = config.app.jobs[key]
        if (jobConfig.hasOwnProperty(POD_TO_RUN)) {
            if (config.app.podName !== jobConfig[POD_TO_RUN]) {
                if (jobConfig.log === true) {
                    logger.debug('[%s] - skipping because podToRun:%s', key, jobConfig[POD_TO_RUN])
                }
                return
            }
        }
        if (jobConfig.log === true) {
            logger.debug('[%s] - running, interval: %s', key, jobConfig.interval)
        }
        try {
            const jobsToRun = require(jobFile)
            await jobsToRun[key]()
        } catch (error) {
            logger.error('runTheJob(): %s - %s', key, error)
        }
    }
}

function isValidInterval(interval) {
    return (interval && interval.length > 0 && !interval.trim().startsWith('#'))
}

async function updateJobInterval({key, config, jobFile, configModel}) {
    const jobConfig = config.app.jobs[key]
    const newInterval = jobConfig.interval
    let currentJob = jobList.get(key)
    if (typeof currentJob === 'undefined') {
        currentJob = {job: null, interval: null}
        jobList.set(key, currentJob)
    }
    if (currentJob.interval !== newInterval) {
        logger.debug('[%s] - updateJobInterval(): interval changed, %s to %s', key, currentJob.interval, newInterval)
        if (isValidInterval(newInterval)) {
            if (jobConfig.hasOwnProperty(POD_TO_RUN)) { // if the attribute is there, it is a single run job
                await configUtils.writeKeyValue(configModel, `app.jobs.${key}.${POD_TO_RUN}`, config.app.podName, 'string')
            }
            if (currentJob.job === null) { // wasn't running before, need to schedule it
                logger.debug('Scheduling job %s every %s', key, newInterval)
                const theJob = getTheJob(jobFile, config, configModel, key)
                currentJob.job = cron.scheduleJob(key, newInterval, theJob)
            } else { // just need to reschedule
                currentJob.job.reschedule(newInterval)
            }
        } else { // Invalid interval need to stop it if running
            if (currentJob.job) { // Is there a running job?
                logger.debug('[%s] - stopping job', key)
                currentJob.job.cancel()
                currentJob.job = null
            } else {
                logger.debug('[%s] - job already stopped', key)
            }
        }
        currentJob.interval = newInterval
        jobList.set(key, currentJob)
    }
}

async function scheduleJobs(baseDir, config = null, configModel = null, messageBus = null) {
    const jobScheduleFunction = async function (theConfig) {
        const jobFile = path.join(baseDir, JOB_FILE)
        for (const key of Object.keys(theConfig.app.jobs)) {
            await updateJobInterval({key, config: theConfig, configModel, jobFile})
        }
    }
    await jobScheduleFunction(config)
    if (messageBus) {
        messageBus.bus.subscribe(messageBus.TOPICS.CONFIG_UPDATED, async () => {
            logger.debug('scheduleJobs(): saw config change, reviewing jobs')
            await jobScheduleFunction(config)
        }).catch(error => logger.error('scheduleJobs(): CONFIG_UPDATE subscription: %s', error))
    }
}

module.exports = {
    cancelJobs,
    initServer,
    scheduleJobs
}
