'use strict';

const fs = require('fs');
const os = require('os');
const path = require('path');
const {DataSource} = require('loopback-datasource-juggler');
const {promisify} = require('util');
const {boundPromisify} = require('../miscUtils');

const JSON_INDENT = 2; // spaces

function buildResetPath(dirname) {
    let split = path.resolve(dirname).split(path.sep);
    let projectName = split[split.length - 4];
    return path.join(os.tmpdir(), `${projectName}_reset.json`);
}

function cleanupArtifacts(resetPath) {
    return promisify(fs.unlink)(resetPath);
}

function dumpResetCommands(models, resetPath) {
    // Iterates over our models, builds the SQL string to truncate each table,
    // and organizes them in an object, keyed by datasource name. Probably won't
    // be too hard to adapt to non-SQL datasources, as whatever's output by
    // buildDelete() is presumably JSON serializable; you'll also need to alter
    // reset() in this file as appropriate.
    let commands = models.reduce(
        (byDb, model) => {
            // Model isn't backed by a DB
            if (!model.getConnector) return byDb;

            let dbConn = model.getConnector();
            let cName = model.getDataSource().name;
            if (!byDb[cName]) byDb[cName] = [];
            byDb[cName].push(dbConn.buildDelete(model.name).sql);
            return byDb;
        },
        {}
    );

    // Dump the object we just constructed to a file.
    return promisify(fs.writeFile)(
        resetPath,
        JSON.stringify(commands, null, JSON_INDENT)
    );
}

module.exports = (dsConfigs, resetPath) => {
    let _resetCommands;
    let connections = {};

    function configureHooks(env) {
        env.beforeAll(connect);
        env.beforeEach(reset);
        env.afterAll(disconnect);
    }

    // Stash a connection object for each datasource defined in dsConfigs
    function connect() {
        for (let [name, conf] of Object.entries(dsConfigs))
            connections[name] = new DataSource(name, conf).connector;
    }

    // Clean up the connections we opened in connect()
    function disconnect() {
        return Promise.all(
            Object.values(connections).map(conn => conn.disconnect())
        );
    }

    // Iterates over and executes the SQL commands - presumably to truncate each
    // table - found in resetCommands; see: dumpResetCommands(). We do this,
    // rather than just iterate over the models and call deleteAll() on each of
    // them in a beforeEach() hook, is that Jest runs each test file in a
    // separate JS environment (subprocess? not sure), which means the app and
    // its models are uninitialized at the start of every test file, which means
    // we would have to load and unload the app per file, taking at least a
    // couple seconds each time.
    function reset() {
        return Promise.all(Object.entries(connections).map(
            async ([name, conn]) => {
                let commands = (await resetCommands())[name].join('; ');
                let sql = `BEGIN; ${commands}; COMMIT;`;
                return boundPromisify(conn, 'execute')(sql);
            }
        ));
    }

    async function resetCommands() {
        if (!_resetCommands) {
            _resetCommands =
                JSON.parse(await promisify(fs.readFile)(resetPath));
        }
        return _resetCommands;
    }

    return {
        configureHooks
    };
};

Object.assign(module.exports, {
    buildResetPath,
    cleanupArtifacts,
    dumpResetCommands
});
